﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Media.Animation;

namespace Example08_Clockgroup2
{
    /// <summary>
    /// MainWindow.xaml の相互作用ロジック
    /// </summary>
    public partial class MainWindow : Window
    {
        // コンストラクター
        public MainWindow()
        {
            InitializeComponent();
            InitializeWindow();
        }

        private const int DST_X = 183;             // 配置先のLeft(DestinationLeft)
        private const int NUM = 4;                 // 縦横各方向のピースの数
        private const int SIZE = 70;               // 1個のピースの縦横サイズ
        private const int INI_X = -100;            // 初期配置のLeft
        private const int INI_Y = 105;             // 初期配置のTop
        private const int TIME = 1;
        private const int TOTAL = NUM * NUM;       // 全ピースの数
        private readonly int[] ORDER =             // ピースの移動順番(10番ピースから開始)
                { 10, 0, 13, 2, 15, 5, 8, 7, 1, 14, 4, 11, 9, 3, 12, 6 }; 
        private ClockGroup _cgRot, _cgTrX, _cgTrY;
        private Piece[] _piece = new Piece[TOTAL]; // Pieceオブジェクトの配列

        // [Avi]ボタンクリックイベント
        private void btnAvi_Click(object sender, RoutedEventArgs e)
        {
            var filepath = GetAviFilePath();
            if (0 < filepath.Length)
            {
                SaveAsAvi(filepath);
                MessageBox.Show("OK");
            }
        }

        // [開始]ボタンクリックイベント
        private void btnStart_Click(object sender, RoutedEventArgs e)
        {
            _cgRot.Controller.Begin();
            _cgTrX.Controller.Begin();
            _cgTrY.Controller.Begin();
        }

        // [停止]ボタンクリックイベント
        private void btnStop_Click(object sender, RoutedEventArgs e)
        {
            _cgRot.Controller.Stop();
            _cgTrX.Controller.Stop();
            _cgTrY.Controller.Stop();
        }


        // 切り取り画像のBitmapを返すメソッド
        private CroppedBitmap GetCrpBmp(BitmapFrame vframe,
                        int left, int top, int width, int height)
        {
            var cbmp = new CroppedBitmap();
            cbmp.BeginInit();
            cbmp.Source = vframe;
            cbmp.SourceRect = new Int32Rect(left, top, width, height);
            cbmp.EndInit();

            return cbmp;
        }

        // Aviファイル保存パス取得
        private string GetAviFilePath()
        {
            var dlg = new Microsoft.Win32.SaveFileDialog();
            dlg.Filter = "aviファイル|*.avi;*.avi";
            var path = String.Empty;

            if (dlg.ShowDialog() == true)
            {
                path = dlg.FileName;
            }

            return path;
        }

        // ピースのアニメーションを生成するメソッド
        private void CreateAnimation()
        {
            // ParallelTimeline生成
            var ptRot = new ParallelTimeline();
            var ptTrX = new ParallelTimeline();
            var ptTrY = new ParallelTimeline();

            // Transform(回転、移動)の配列生成
            RotateTransform[] rottran = new RotateTransform[TOTAL];
            TranslateTransform[] tranX = new TranslateTransform[TOTAL];
            TranslateTransform[] tranY = new TranslateTransform[TOTAL];

            var tsDur = new TimeSpan(0, 0, TIME);　　// アニメーションのTimeSpan

            // EasingFunction設定
            var easing = new CircleEase();
            easing.EasingMode = EasingMode.EaseOut;

            // ピースのアニメーションとClockGroup生成
            for (int i = 0; i < TOTAL; i++)
            {
                var tsBegin = new TimeSpan(0, 0, i * TIME + 1);

                // Transform(回転、移動)設定
                rottran[i] = new RotateTransform();
                rottran[i].CenterX = SIZE / 2;
                rottran[i].CenterY = SIZE / 2;
                tranX[i] = new TranslateTransform();
                tranY[i] = new TranslateTransform();
                var groupT = new TransformGroup();
                groupT.Children.Add(rottran[i]);
                groupT.Children.Add(tranX[i]);
                groupT.Children.Add(tranY[i]);
                _piece[i].RenderTransform = groupT;

                // ピース回転のアニメーション
                var daRot = new DoubleAnimation(720, new Duration(tsDur));
                daRot.BeginTime = tsBegin;
                daRot.EasingFunction = easing;
                ptRot.Children.Add(daRot);

                // ピース横方向移動のアニメーション
                var daX = new DoubleAnimation(_piece[i].DestX, new Duration(tsDur));
                daX.BeginTime = tsBegin;
                daX.EasingFunction = easing;
                ptTrX.Children.Add(daX);

                // ピース縦方向移動のアニメーション
                var daY = new DoubleAnimation(_piece[i].DestY, new Duration(tsDur));
                daY.BeginTime = tsBegin;
                daY.EasingFunction = easing;
                ptTrY.Children.Add(daY);
            }

            // ClockGroup生成
            _cgRot = ptRot.CreateClock();    // 回転のClockGroup
            _cgTrX = ptTrX.CreateClock();    // 横方向移動のClockGroup
            _cgTrY = ptTrY.CreateClock();    // 縦方向移動のClockGroup

            // Transform(回転、移動)へClockGroupの子要素適用
            for (int i2 = 0; i2 < TOTAL; i2++)
            {
                rottran[i2].ApplyAnimationClock(RotateTransform.AngleProperty,
                                               (AnimationClock)_cgRot.Children[i2]);
                tranX[i2].ApplyAnimationClock(TranslateTransform.XProperty,
                                             (AnimationClock)_cgTrX.Children[i2]);
                tranY[i2].ApplyAnimationClock(TranslateTransform.YProperty,
                                             (AnimationClock)_cgTrY.Children[i2]);
            }

            _cgRot.Controller.Stop();
            _cgTrX.Controller.Stop();
            _cgTrY.Controller.Stop();
        }
        
        // ピースを生成するメソッド
        private void CreatePieces()
        {
            var frame = BitmapFrame.Create(new Uri("Pic1.JPG", UriKind.Relative));
            
            for (int i = 0; i < TOTAL; i++)
            {
                _piece[i] = new Piece();  // Pieceクラス(最下行で定義)のインスタンス生成

                Canvas.SetLeft(_piece[i], INI_X);
                Canvas.SetTop(_piece[i], INI_Y);
                var x = ORDER[i] / NUM * SIZE;
                var y = ORDER[i] % NUM * SIZE;
                _piece[i].DestX = DST_X + x;
                _piece[i].DestY = y - INI_Y;
                _piece[i].Source = GetCrpBmp(frame, x, y, SIZE, SIZE);
                cvs1.Children.Add(_piece[i]);
            }
        }

        // ピースとそのアニメーションの設定
        private void InitializeWindow()
        {
            CreatePieces();
            CreateAnimation();
        }

        // Aviファイル作成
        private void SaveAsAvi(string filepath)
        {
            var fps = 48;                                  // フレームレート
            var totalFrames = fps * TIME * (TOTAL + 2);    // 全フレーム数
            var secs = Enumerable.Range(0, totalFrames).Select(t => (((double)t) / fps));
            var aviManager = new AviFile.AviManager(filepath, false);
            AviFile.VideoStream aviStream = null;
            var tempFile = "frame.png";

            foreach (var sec in secs)
            {
                _cgRot.Controller.SeekAlignedToLastTick(TimeSpan.FromSeconds(sec),
                                                        TimeSeekOrigin.BeginTime);
                _cgTrX.Controller.SeekAlignedToLastTick(TimeSpan.FromSeconds(sec),
                                                        TimeSeekOrigin.BeginTime);
                _cgTrY.Controller.SeekAlignedToLastTick(TimeSpan.FromSeconds(sec),
                                                        TimeSeekOrigin.BeginTime);
                cvs1.UpdateLayout();
                cvs1.SaveImage(tempFile);        // cvs1の画像をファイルに保存する
                var bmp = new System.Drawing.Bitmap(tempFile);
                if (aviStream == null)
                    aviStream = aviManager.AddVideoStream(true, fps, bmp);
                else
                    aviStream.AddFrame(bmp);
                bmp.Dispose();
            }
            aviManager.Close();

            _cgRot.Controller.Stop();
            _cgTrX.Controller.Stop();
            _cgTrY.Controller.Stop();
            
            System.IO.File.Delete(tempFile);
        }


        // Pieceクラスの定義－ImageクラスにDestX, DestYプロパティを追加したもの
        private class Piece : Image
        {
            public int DestX { get; set; }   // ピース着地点のX座標
            public int DestY { get; set; }   // ピース着地点のY座標
        }

    }

    public static class CanvasExtension
    {
        /// <summary> [SaveImage]：Canvasの画像をファイルに保存するCanvasクラスの拡張メソッド 
        /// 出典：http://qiita.com/hugo-sb/items/894914f6bbe224a45d49
        /// </summary>
        /// <param name="canvas"></param>
        /// <param name="path">保存する画像ファイル名</param>
        /// <param name="encoder">BitmapEncoderが指定されなかった場合はPNG形式になる</param>
        public static void SaveImage(this Canvas canvas, string path, BitmapEncoder encoder = null)
        {
            // レイアウトを再計算させる
            var size = new Size(canvas.ActualWidth, canvas.ActualHeight);
            canvas.Measure(size);
            canvas.Arrange(new Rect(size));

            // VisualObjectをBitmapに変換する
            var renderBitmap = new RenderTargetBitmap((int)size.Width,       // 画像の幅
                                                      (int)size.Height,      // 画像の高さ
                                                      96.0d,                 // 横96.0DPI
                                                      96.0d,                 // 縦96.0DPI
                                                      PixelFormats.Pbgra32); // 32bit(RGBA各8bit)
            renderBitmap.Render(canvas);

            // 出力用の FileStream を作成する
            using (var stream = new System.IO.FileStream(path, System.IO.FileMode.Create))
            {
                // 変換したBitmapをエンコードしてFileStreamに保存する。
                // BitmapEncoder が指定されなかった場合は、PNG形式とする。
                encoder = encoder ?? new PngBitmapEncoder();
                encoder.Frames.Add(BitmapFrame.Create(renderBitmap));
                encoder.Save(stream);
            }
        }

    }

}
